part of circumstellar;

/// A list of static endpoints and resources for API connections.
class DiscordGatewayStatic {
  /// Whether Discord supports connecting with a specific API version.
  static Map<int, bool> gatewayCompatibility = {
    3: false,
    4: false,
    5: false,
    6: true
  };

  /// Base URL for the REST API without a version specified.
  static String restApiBaseNoVer = "https://discordapp.com/api";

  /// The current gateway version supported.
  static int gatewayVersion = 6;
  // Builds an API base string using the default (or a specified) version.
  static String getRestApiBase([int version = -1]) {
    int gwVer = (version == -1 ? DiscordGatewayStatic.gatewayVersion : version);
    return (DiscordGatewayStatic.restApiBaseNoVer + "/v${gwVer}");
  }

  /// A map of error codes to descriptions for a WebSocket close event.
  static Map<int, String> gatewayCloseEventCodes = {
    4000: 'Unknown error',
    4001: 'Unknown opcode',
    4002: 'Decode error',
    4003: 'Not authenticated',
    4004: 'Authentication failed',
    4005: 'Already authenticated',
    4007: 'Invalid sequence (during resume)',
    4008: 'Rate limited',
    4009: 'Session timeout',
    4010: 'Invalid shard',
    4011: 'Sharding required',
  };

  /// A map of all OP codes, their names and descriptions.
  static Map<int, OPCodeDescription> opCodes = {
    // clientReceives first
    // clientSends second
    0: new OPCodeDescription(
        0, 'Dispatch', 'DISPATCH', true, false, 'Dispatches an event'),
    1: new OPCodeDescription(
        1, 'Heartbeat', 'HEARTBEAT', true, true, 'Used for ping checking'),
    2: new OPCodeDescription(
        2, 'Identify', 'IDENTIFY', false, true, 'Used for client handshake'),
    3: new OPCodeDescription(3, 'Status Update', 'STATUS UPDATE', false, true,
        'Used to update the clientstatus'),
    4: new OPCodeDescription(4, 'Voice State Update', 'VOICE_STATE_UPDATE',
        false, true, 'Used to join/move/leave voice channels'),
    5: new OPCodeDescription(5, 'Voice Server Ping', 'VOICE_SERVER_PING', false,
        true, 'Used for voice ping checking'),
    6: new OPCodeDescription(6, 'Resume', 'RESUME', true, false,
        'Used to resume a closed connection'),
    7: new OPCodeDescription(7, 'Reconnect', 'RECONNECT', true, false,
        'Used to tell clients to reconnect to the gateway'),
    8: new OPCodeDescription(8, 'Request Guild Members',
        'REQUEST_GUILD_MEMBERS', false, true, 'Used to request guild members'),
    9: new OPCodeDescription(9, 'Invalid Session', 'INVALID_SESSION', true,
        false, 'Used to notify client they have an invalid session ID'),
    10: new OPCodeDescription(10, 'Hello', 'HELLO', true, false,
        'sent immediately after connecting, contains heartbeat and server debug information'),
    11: new OPCodeDescription(11, 'Heartbeat ACK', 'HEARTBEAT_ACK', true, false,
        'Sent immediately following a client heartbeat that was received')
  };
}

/// Represents a route/endpoint that the client can request from Discord.
class ApiRoute {
  /// Represents the route that this endpoint builds on. It will be null if it is the base URI.
  ApiRoute baseRoute;

  /// The endpoint that this route is connecting to.
  String endpoint;

  /// The method that should be used to call this endpoint.
  String method;

  /// Creates an [ApiRoute].
  ApiRoute(this.endpoint, [this.method = 'GET', this.baseRoute]);

  /// Compiles a [String] URL to use for requesting.
  String buildRouteUrl() {
    if (baseRoute == null) return endpoint;
    return baseRoute.buildRouteUrl() + endpoint;
  }
}

/// A static list of all available API endpoints.
class ApiRoutes {
  /// Represents the base route of all Discord REST connections.
  static ApiRoute baseRoute =
      new ApiRoute(DiscordGatewayStatic.getRestApiBase());

  /// Represents the [/users] Discord endpoint.
  static UsersRoute users = new UsersRoute();

  /// Represents the [/gateway] Discord endpoint.
  static GatewayRoute gateway = new GatewayRoute();

  /// Represents the [/channels] Discord endpoint.
  static ChannelsRoute channels = new ChannelsRoute();

  /// Represents the [/guilds] Discord endpoint.
  static GuildsRoute guilds = new GuildsRoute();
}

/// Represents the [/channels] Discord endpoint.
class ChannelsRoute {
  /// The base route for all [/channels] endpoints.
  ApiRoute channelsEndpoint = new ApiRoute('/channels')
    ..baseRoute = ApiRoutes.baseRoute;

  /// Gets a channel by ID.
  ///
  /// Route: GET [/channels/{channel.id}]
  ///
  /// Returns a channel object.
  ApiRoute getChannelWithId(String id) =>
      new ApiRoute('/${id}', 'GET', channelsEndpoint);

  /// Update a channels settings.
  ///
  /// Route: PUT/PATCH [/channels/{channel.id}]
  ///
  /// Requires the 'MANAGE_CHANNELS' permission for the guild.
  /// Returns a channel on success, and a 400 BAD REQUEST on invalid parameters.
  /// Fires a Channel Update Gateway event If modifying a category,
  /// individual Channel Update events will fire for each child channel that also changes.
  /// For the PATCH method, all the JSON Params are optional.
  ApiRoute modifyChannelWithId(String id, String method) =>
      new ApiRoute('/${id}', method, channelsEndpoint);

  /// Deletes a channel, or closes a private message.
  ///
  /// Route: DELETE [/channels/{channel.id}]
  ///
  /// Requires the 'MANAGE_CHANNELS' permission for the guild.
  /// Deleting a category does not delete its child channels; they will have their parent_id removed and a Channel Update Gateway event will fire for each of them.
  /// Returns a channel object on success. Fires a Channel Delete Gateway event.
  ApiRoute deleteChannelWithId(String id) =>
      new ApiRoute('/${id}', 'DELETE', channelsEndpoint);

  // Messages
  ApiRoute getChannelMessages(String id) =>
      new ApiRoute('/${id}/messages', 'GET', channelsEndpoint);
  ApiRoute getChannelMessage(String channelId, String messageId) =>
      new ApiRoute(
          '/${channelId}/messages/${messageId}', 'GET', channelsEndpoint);
  ApiRoute createMessage(String channelId) =>
      new ApiRoute('/${channelId}/messages', 'POST', channelsEndpoint);

  // Reactions
  ApiRoute createReaction(String channelId, String messageId, String emoji) =>
      new ApiRoute('/${channelId}/messages/${messageId}/reactions/${emoji}/me',
          'PUT', channelsEndpoint);
  ApiRoute deleteOwnReaction(
          String channelId, String messageId, String emoji) =>
      new ApiRoute('/${channelId}/messages/${messageId}/reactions/${emoji}/me',
          'DELETE', channelsEndpoint);
  ApiRoute deleteUserReaction(
          String channelId, String messageId, String emoji, String userId) =>
      new ApiRoute(
          '/${channelId}/messages/${messageId}/reactions/${emoji}/${userId}',
          'DELETE',
          channelsEndpoint);
  // TODO: Finish
}

class UsersRoute {
  ApiRoute usersEndpoint = new ApiRoute('/users')
    ..baseRoute = ApiRoutes.baseRoute;

  ApiRoute getCurrentUser() => new ApiRoute('/@me', 'GET', usersEndpoint);
  ApiRoute getUserWithId(String id) =>
      new ApiRoute('/${id}', 'GET', usersEndpoint);
  ApiRoute modifyCurrentUser() => new ApiRoute('/@me', 'PATCH', usersEndpoint);
  ApiRoute getCurrentUserGuilds() =>
      new ApiRoute('/@me/guilds', 'GET', usersEndpoint);
  ApiRoute leaveGuild(String guildId) =>
      new ApiRoute('/@me/guilds/${guildId}', 'DELETE', usersEndpoint);
  ApiRoute getUserDms() => new ApiRoute('/@me/channels', 'GET', usersEndpoint);
  ApiRoute createDmOrGroupDm() =>
      new ApiRoute('/@me/channels', 'POST', usersEndpoint);
  ApiRoute getUserConnections() =>
      new ApiRoute('/@me/connections', 'GET', usersEndpoint);
}

class GatewayRoute {
  ApiRoute gatewayEndpoint = new ApiRoute('/gateway')
    ..baseRoute = ApiRoutes.baseRoute;

  ApiRoute getGateway() => gatewayEndpoint;
  ApiRoute getGatewayBot() => new ApiRoute('/bot', 'GET', gatewayEndpoint);
}

class GuildsRoute {
  ApiRoute guildsEndpoint = new ApiRoute('/guilds')
    ..baseRoute = ApiRoutes.baseRoute;

  /// Gets a guild by ID.
  ///
  /// Route: GET [/guilds/{guild.id}]
  ApiRoute getGuildWithId(String id) =>
      new ApiRoute('/$id', 'GET', guildsEndpoint);

  /// Modifies a guild's settings.
  ///
  /// Route: PATCH [/guilds/{guild.id}]
  ///
  /// Requires the 'MANAGE_GUILD' permission.
  /// Returns the updated guild object on success.
  /// Fires a Guild Update Gateway event.
  ApiRoute modifyGuild(String id) =>
      new ApiRoute('/$id', 'PATCH', guildsEndpoint);

  /// Creates a new guild.
  ///
  /// Route: POST [/guilds]
  ///
  /// Returns a guild object on success.
  /// Fires a Guild Create Gateway event.
  ApiRoute createGuild() => new ApiRoute('', 'POST', guildsEndpoint);

  /// Deletes a guild permanently.
  ///
  /// Route: DELETE /guilds/{guild.id}
  ///
  /// User must be owner. Returns [204 No Content] on success.
  /// Fires a Guild Delete Gateway event.
  ApiRoute deleteGuild(String id) =>
      new ApiRoute('/$id', 'DELETE', guildsEndpoint);
}

/// Describes any OP code.
class OPCodeDescription {
  int code;
  String friendlyName;
  String serverName;
  bool clientReceives;
  bool clientSends;
  String description;
  OPCodeDescription(this.code, this.friendlyName, this.serverName,
      this.clientReceives, this.clientSends, this.description);
}

/// Enumerable of all channel types.
class ChannelType {
  static int GuildText = 0;
  static int Dm = 1;
  static int GuildVoice = 2;
  static int GroupDm = 3;
  static int GuildCategory = 4;
}

/// Static methods to help with resolving channel types.
class ChannelTypeHelper {
  /// Resolves a channel type from an integer.
  static String getTypeFromInt(int type) {
    if (type == 0) return 'GuildTextChannel';
    if (type == 1) return 'DMChannel';
    if (type == 2) return 'GuildVoiceChannel';
    if (type == 3) return 'GroupDMChannel';
    if (type == 4) return 'GuildCategoryChannel';
    return null;
  }

  /// Resolves a channel type from a [BaseChannel] object.
  static String getTypeFromChannel(BaseChannel channel) =>
      ChannelTypeHelper.getTypeFromInt(channel.type);
}

/// Enumerable of all message types.
class MessageType {
  static int Default = 0;
  static int RecipientAdd = 1;
  static int RecipientRemove = 2;
  static int Call = 3;
  static int ChannelNameChange = 4;
  static int ChannelIconChange = 5;
  static int ChannelPinnedMessage = 6;
  static int GuildMemberJoin = 7;
}

/// Static methods to help with resolving message types.
class MessageTypeHelper {
  /// Resolves the type of a message from an integer.
  static String getTypeFromInt(int type) {
    if (type == MessageType.Default) return 'DefaultMessage';
    if (type == MessageType.RecipientAdd) return 'RecipientAddMessage';
    if (type == MessageType.RecipientRemove) return 'RecipientRemoveMessage';
    if (type == MessageType.Call) return 'CallMessage';
    if (type == MessageType.ChannelNameChange)
      return 'ChannelNameChangeMessage';
    if (type == MessageType.ChannelIconChange)
      return 'ChannelIconChangeMessage';
    if (type == MessageType.ChannelPinnedMessage) return 'ChannelPinnedMessage';
    if (type == MessageType.GuildMemberJoin) return 'GuildMemberJoinMessage';
    return null;
  }

  /// Resolves the type of a message from a [Message] object.
  static String getTypeFromMessage(Message message) =>
      getTypeFromInt(message.type);
}

/// Enumerable for guild verification level types.
class GuildVerificationLevel {
  /// Unrestricted
  static int None = 0;

  /// Must have a verified email on account
  static int Low = 1;

  /// Must be registered on Discord for longer than 5 minutes
  static int Medium = 2;

  /// Must be a member of this server for longer than 10 minutes
  static int High = 3;

  /// Must have a verified phone number
  static int VeryHigh = 4;
}

/// Static methods to help with resolving guild verification types.
class GuildVerificationLevelHelper {
  /// Resolves a verification level from an integer.
  static String getLevelFromInt(int type) {
    if (type == GuildVerificationLevel.None) return 'Unrestricted';
    if (type == GuildVerificationLevel.Low) return 'Low';
    if (type == GuildVerificationLevel.Medium) return 'Medium';
    if (type == GuildVerificationLevel.High) return 'High';
    if (type == GuildVerificationLevel.VeryHigh) return 'Very High';
    return null;
  }

  /// Resolves a verification level from a [Guild] object.
  static String getLevelFromGuild(Guild guild) =>
      getLevelFromInt(guild.verificationLevel);
}

/// Enumerable for MFA level types.
class MFALevel {
  static int None = 0;
  static int Elevated = 1;
}

/// Static methods to help with resolving MFA level types.
class MFALevelHelper {
  /// Resolves an MFA level from an integer.
  static String getLevelFromInt(int type) {
    if (type == MFALevel.None) return 'None';
    if (type == MFALevel.Elevated) return 'Elevated';
    return null;
  }

  /// Resolves an MFA level from a [Guild] object.
  static String getLevelFromGuild(Guild guild) =>
      getLevelFromInt(guild.mfaLevel);
}

/// Enumerable for explicit content filter level types.
class ExplicitContentFilterLevel {
  static int Disabled = 0;
  static int MembersWithoutRoles = 1;
  static int AllMembers = 2;
}

/// Static methods to help with resolving Explicit Content Filter types.
class ExplicitContentFilterLevelHelper {
  /// Resolves an ECF level from an integer.
  static String getLevelFromInt(int type) {
    if (type == ExplicitContentFilterLevel.Disabled) return 'Disabled';
    if (type == ExplicitContentFilterLevel.MembersWithoutRoles)
      return 'Members Without Roles';
    if (type == ExplicitContentFilterLevel.AllMembers) return 'All Members';
    return null;
  }

  /// Resolves an ECF level from a [Guild] object.
  static String getLevelFromGuild(Guild guild) =>
      getLevelFromInt(guild.explicitContentFilterLevel);
}

/// Enumerable for default message notification level types.
class DefaultMessageNotificationLevel {
  static int AllMessages = 0;
  static int OnlyMentions = 1;
}

/// Static methods to help with resolving Default Message Notification types.
class DefaultMessageNotificationLevelHelper {
  /// Resolves a DMN level from an integer.
  static String getLevelFromInt(int type) {
    if (type == DefaultMessageNotificationLevel.AllMessages)
      return 'All Messages';
    if (type == DefaultMessageNotificationLevel.OnlyMentions)
      return 'Only Mentions';
    return null;
  }

  /// Resolves a DMN level from a [Guild] object.
  static String getLevelFromGuild(Guild guild) =>
      getLevelFromInt(guild.defaultMessageNotificationLevel);
}
