part of circumstellar;

/// Useful logging utility. See the Circumstellar Logging wiki page for more information.
class LoggerContext {
  /// The parent of this [LoggerContext].
  LoggerContext parent = null;

  /// The default logging implementation.
  LoggingImplementation _default = null;

  /// The name of this logger, used by the default implementation.
  String name;

  /// Initializes the default logging implementation.
  void _internalInit() {
    this._default = new DefaultLoggingImplementation(this);
    this._loggingImplementations.add(_default);
  }

  /// Creates a new [LoggerContext] as a child of another [LoggerContext].
  LoggerContext.fromParent(this.parent, this.name) {
    this._internalInit();
  }

  /// Creates a new root [LoggerContext] with no parent.
  LoggerContext.root(this.name) {
    this._internalInit();
  }

  /// Adds a [LoggingImplementation] to the internal list.
  void addImplementation(LoggingImplementation impl) {
    this._loggingImplementations.add(impl);
  }

  /// Removes a [LoggingImplementation] to the internal list.
  void removeImplementation(LoggingImplementation impl) {
    this._loggingImplementations.remove(impl);
  }

  /// Removes the [DefaultLoggingImplementation] from the internal list.
  void removeDefaultImplementation() {
    this.removeImplementation(this._default);
  }

  /// A list of all registered [LoggingImplementation]s.
  List<LoggingImplementation> _loggingImplementations = [];

  /// Logs a message to the standard output stream.
  void log(LogLevel level, String comment) {
    _loggingImplementations.forEach((impl) {
      impl.log(level, comment);
    });
  }
}

abstract class LoggingImplementation {
  LoggerContext context;
  LoggingImplementation(this.context);
  void log(LogLevel level, String comment);
}

class DefaultLoggingImplementation implements LoggingImplementation {
  String logIndent = '    ';
  String buildLoggerTree() {
    if (context.parent == null) {
      return context.name;
    } else
      return this.buildLoggerTree() + ' -> ' + context.name;
  }

  LoggerContext context;

  DefaultLoggingImplementation(this.context);

  @override
  void log(LogLevel level, String comment) {
    print(
        '${level.toString().replaceAll('LogLevel.', '')}: [${this.buildLoggerTree()}] at ${new DateTime.now().toLocal().toString()}');
    print(this.logIndent + comment);
  }
}

enum LogLevel { Trace, Debug, Info, Warn, Error, Critical, Crash }
